@@ -32,12 +32,14 @@
 #include <gconf/gconf-client.h>
 #include <nautilus-burn-drive.h>
 #include <libgnome/gnome-help.h>
+#include <libgnome/gnome-url.h>
 #include <libgnomeui/gnome-ui-init.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <profiles/gnome-media-profiles.h>
 #include <gst/gst.h>
 
 #include "bacon-message-connection.h"
+#include "gconf-bridge.h"
 #include "sj-about.h"
 #include "sj-metadata.h"
 #include "sj-metadata-musicbrainz.h"
@@ -70,6 +72,7 @@ static GtkWidget *title_entry, *artist_entry, *duration_label, *genre_entry;
 static GtkWidget *track_listview, *extract_button, *play_button;
 static GtkWidget *status_bar;
 static GtkWidget *extract_menuitem, *play_menuitem, *next_menuitem, *prev_menuitem, *select_all_menuitem, *deselect_all_menuitem;
+static GtkWidget *submit_menuitem;
 GtkListStore *track_store;
 static BaconMessageConnection *connection;
 GtkCellRenderer *toggle_renderer, *title_renderer, *artist_renderer;
@@ -79,14 +82,17 @@ NautilusBurnDrive *drive = NULL;
 gboolean strip_chars;
 gboolean eject_finished;
 gboolean extracting = FALSE;
-static gboolean tray_opened = FALSE;
-static guint poll_id = 0;
+
 static gint total_no_of_tracks;
 static gint no_of_tracks_selected;
 static AlbumDetails *current_album;
+static char *current_submit_url = NULL;
 
 gboolean autostart = FALSE, autoplay = FALSE;
 
+/* For the default drive stuff, TODO: remove */
+static GList *cdroms = NULL;
+
 #define DEFAULT_PARANOIA 4
 #define RAISE_WINDOW "raise-window"
 #define SOURCE_GLADE "../data/sound-juicer.glade"
@@ -152,24 +158,6 @@ sj_stock_init (void)
   initialized = TRUE;
 }
 
-static void
-set_busy_cursor (gboolean busy)
-{
-  if (!GTK_WIDGET_REALIZED (main_window))
-    return;
-
-  if (busy) {
-    GdkCursor *cursor;
-    cursor = gdk_cursor_new_for_display (gdk_drawable_get_display (main_window->window),
-					 GDK_WATCH);
-    gdk_window_set_cursor (main_window->window, cursor);
-    gdk_cursor_unref (cursor);
-    gdk_display_sync (gdk_drawable_get_display (main_window->window));
-  } else {
-    gdk_window_set_cursor (main_window->window, NULL);
-  }
-}
-
 void
 sj_main_set_title (const char* detail)
 {
@@ -226,33 +214,6 @@ void on_eject_activate (GtkMenuItem *item, gpointer user_data)
   nautilus_burn_drive_eject (drive);
 }
 
-static gboolean poll_helper (gpointer data)
-{
-  reread_cd (TRUE);
-  return FALSE;
-}
-
-static gboolean poll_tray_opened (gpointer data)
-{
-  gboolean new_status;
-  if (extracting == TRUE)
-    return TRUE;
-
-  new_status = nautilus_burn_drive_door_is_open (drive);
-  if (new_status != tray_opened && new_status == FALSE) {
-    d(g_printerr("Poll detected new CD\n"));
-    set_busy_cursor (TRUE);
-    g_timeout_add (300, poll_helper, NULL);
-  } else if (new_status != tray_opened && new_status == TRUE) {
-    d(g_printerr("Poll detected removed CD\n"));
-    stop_ui_hack ();
-    update_ui_for_album (NULL);
-  }
-  tray_opened = new_status;
-
-  return TRUE;
-}
-
 gboolean on_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data)
 {
   if (extracting) {
@@ -269,9 +230,6 @@ gboolean on_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data
     gtk_widget_destroy (dialog);
     
     if (response == GTK_RESPONSE_ACCEPT) {
-      if (poll_id > 0) {
-        g_source_remove (poll_id);
-      }
       return FALSE;
     } 
     return TRUE;
@@ -603,10 +561,11 @@ static void
 metadata_cb (SjMetadata *m, GList *albums, GError *error)
 {
   gboolean realized = GTK_WIDGET_REALIZED (main_window);
-  
-  set_busy_cursor (FALSE);
-  /* Clear the statusbar message */
-  gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 0);
+
+  if (realized)
+    gdk_window_set_cursor (main_window->window, NULL);
+    /* Clear the statusbar message */
+    gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 0);
   
   if (error && !(error->code == SJ_ERROR_CD_NO_MEDIA)) {
     GtkWidget *dialog;
@@ -663,9 +622,16 @@ metadata_cb (SjMetadata *m, GList *albums, GError *error)
 static gboolean
 is_audio_cd (NautilusBurnDrive *drive)
 {
+  NautilusBurnMediaType type;
   gboolean audio;
   if (drive == NULL) return FALSE;
-  nautilus_burn_drive_get_media_type_full (drive, NULL, NULL, NULL, &audio);
+  type = nautilus_burn_drive_get_media_type_full (drive, NULL, NULL, NULL, &audio);
+  if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
+    g_warning ("Error getting media type\n");
+  }
+  if (type == NAUTILUS_BURN_MEDIA_TYPE_BUSY) {
+    g_warning ("BUSY getting media type, should re-check\n");
+  }
   return audio;
 }
 
@@ -675,25 +641,42 @@ is_audio_cd (NautilusBurnDrive *drive)
 static void reread_cd (gboolean ignore_no_media)
 {
   /* TODO: remove ignore_no_media? */
-  gboolean realized = GTK_WIDGET_REALIZED (main_window);
   GError *error = NULL;
+  GdkCursor *cursor;
+  gboolean realized = GTK_WIDGET_REALIZED (main_window);
   
   /* Set watch cursor */
-  set_busy_cursor (TRUE);
+  if (realized) {
+    cursor = gdk_cursor_new_for_display (gdk_drawable_get_display (main_window->window),
+					 GDK_WATCH);
+    gdk_window_set_cursor (main_window->window, cursor);
+    gdk_cursor_unref (cursor);
+    gdk_display_sync (gdk_drawable_get_display (main_window->window));
+  }
   
   /* Set statusbar message */
   gtk_statusbar_push(GTK_STATUSBAR(status_bar), 0, _("Retrieving track listing...please wait."));
 
   d(if (!drive) g_printerr("Attempting to re-read NULL drive\n"));
 
+  g_free (current_submit_url);
+  current_submit_url = NULL;
+  gtk_widget_set_sensitive (submit_menuitem, FALSE);
+
   if (!is_audio_cd (drive)) {
     d(g_printerr("Media is not an audio CD\n"));
     update_ui_for_album (NULL);
     gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 0);
-    set_busy_cursor (FALSE);
+    if (realized)
+      gdk_window_set_cursor (main_window->window, NULL);
     return;
   }
-  
+
+  current_submit_url = sj_metadata_get_submit_url (metadata);
+  if (current_submit_url) {
+    gtk_widget_set_sensitive (submit_menuitem, TRUE);
+  }
+
   sj_metadata_list_albums (metadata, &error);
 
   if (error && !(error->code == SJ_ERROR_CD_NO_MEDIA && ignore_no_media)) {
@@ -721,10 +704,66 @@ static void reread_cd (gboolean ignore_no_media)
 }
 
 static void
-set_device(const char* device, gboolean ignore_no_media)
+media_added_cb (NautilusBurnDrive *drive,
+                gpointer           data)
 {
-  if (device == NULL) {
+  if (extracting == TRUE) {
+    /* FIXME: recover? */
+  }
+
+  d(g_printerr ("** media added to device %s\n", drive->device));
+  reread_cd (TRUE);
+}
+
+static void
+media_removed_cb (NautilusBurnDrive *drive,
+                  gpointer           data)
+{
+  if (extracting == TRUE) {
+    /* FIXME: recover? */
+  }
+
+  d(g_printerr ("** media removed from device %s\n", drive->device));
+  stop_ui_hack ();
+  update_ui_for_album (NULL);
+}
+
+static void
+set_drive_from_device (const char *device)
+{
+  gboolean is_locked;
+  char    *reason;
+
+  if (drive) {
+    nautilus_burn_drive_unlock (drive);
+    nautilus_burn_drive_unref (drive);
     drive = NULL;
+  }
+
+  if (! device)
+    return;
+
+  drive = nautilus_burn_drive_new_from_path (device);
+  g_assert (drive);
+
+  is_locked = nautilus_burn_drive_lock (drive, _("Extracting audio from CD"), &reason);
+  if (! is_locked) {
+    g_warning ("Could not lock drive: %s", reason);
+    g_free (reason);
+  }
+
+  g_object_set (drive, "enable-monitor", TRUE, NULL);
+  g_signal_connect (drive, "media-added", G_CALLBACK (media_added_cb), NULL);
+  g_signal_connect (drive, "media-removed", G_CALLBACK (media_removed_cb), NULL);
+}
+
+static void
+set_device (const char* device, gboolean ignore_no_media)
+{
+  gboolean tray_opened;
+
+  if (device == NULL) {
+    set_drive_from_device (device);
   } else if (access (device, R_OK) != 0) {
     GtkWidget *dialog;
     char *message;
@@ -742,21 +781,51 @@ set_device(const char* device, gboolean ignore_no_media)
     gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), TRUE);
     gtk_dialog_run (GTK_DIALOG (dialog));
     gtk_widget_destroy (dialog);
+
     /* Set a null device */
-    drive = NULL;
+    set_drive_from_device (NULL);
   } else {
-    drive = nautilus_burn_drive_new_from_path (device);
-    g_assert (drive);
+    set_drive_from_device (device);
   }
+
   sj_metadata_set_cdrom (metadata, device);
   sj_extractor_set_device (extractor, device);
   
-  tray_opened = nautilus_burn_drive_door_is_open (drive);
-  if (tray_opened == FALSE) {
-    reread_cd (ignore_no_media);
+  if (drive != NULL) {
+    tray_opened = nautilus_burn_drive_door_is_open (drive);
+    if (tray_opened == FALSE) {
+      reread_cd (ignore_no_media);
+    }
   }
 }
 
+gboolean cd_drive_exists (const char *device)
+{
+  GList *l;
+  if (device == NULL) return FALSE;
+  if (cdroms == NULL) {
+    cdroms = nautilus_burn_drive_get_list (FALSE, FALSE);
+  }
+  for (l = cdroms; l != NULL; l = l->next) {
+    if (strcmp (((NautilusBurnDrive *) (l->data))->device, device) == 0)
+      return TRUE;
+  }
+  return FALSE;
+}
+
+const char* prefs_get_default_device ()
+{
+  static const char* default_device = NULL;
+  if (default_device == NULL) {
+    NautilusBurnDrive *cd;
+    cdroms = nautilus_burn_drive_get_list (FALSE, FALSE);
+    if (cdroms == NULL) return NULL;
+    cd = cdroms->data;
+    default_device = cd->device;
+  }
+  return default_device;
+}
+
 /**
  * The GConf key for the device changed
  */
@@ -785,7 +854,6 @@ void device_changed_cb (GConfClient *client, guint cnxn_id, GConfEntry *entry, g
       exit (1);
 #endif
     }
-    drive = nautilus_burn_drive_new_from_path (device);
   } else {
     device = gconf_value_get_string (entry->value);
   }
@@ -881,6 +949,35 @@ void on_reread_activate (GtkWidget *button, gpointer user_data)
   reread_cd (FALSE);
 }
 
+/**
+ * Clicked the Submit menu item in the UI
+ */
+void on_submit_activate (GtkWidget *menuitem, gpointer user_data)
+{
+  GError *error = NULL;
+
+  if (current_submit_url) {
+    if (!gnome_url_show (current_submit_url, &error)) {
+      GtkWidget *dialog;
+
+      dialog = gtk_message_dialog_new (GTK_WINDOW (main_window),
+                                       GTK_DIALOG_DESTROY_WITH_PARENT,
+                                       GTK_MESSAGE_ERROR,
+                                       GTK_BUTTONS_CLOSE,
+                                       "<b>%s</b>\n\n%s\n%s: %s",
+				       _("Could not open URL"),
+				       _("Sound Juicer could not open the submission URL"),
+				       _("Reason"),
+				       error->message);
+      gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), TRUE);
+      gtk_dialog_run (GTK_DIALOG (dialog));
+      gtk_widget_destroy (dialog);
+      g_error_free (error);
+    }
+  }
+
+}
+
 /**
  * Called in on_extract_toggled to see if there are selected tracks or not.
  * extracting points to the boolean to set to if there are tracks to extract,
@@ -1049,6 +1146,7 @@ static const char* genres[] = {
   "Dance",
   "Electronica",
   "Folk",
+  "Funk",
   "Jazz",
   "Latin",
   "Pop",
@@ -1141,6 +1239,8 @@ int main (int argc, char **argv)
 
   sj_stock_init ();
 
+  gtk_window_set_default_icon_from_file (PIXMAPDIR"/sound-juicer.png", NULL);
+
   connection = bacon_message_connection_new ("sound-juicer");
   if (bacon_message_connection_get_is_server (connection) == FALSE) {
 	  bacon_message_connection_send (connection, RAISE_WINDOW);
@@ -1198,10 +1298,9 @@ int main (int argc, char **argv)
   glade_xml_signal_autoconnect (glade);
 
   main_window = glade_xml_get_widget (glade, "main_window");
-  gtk_window_set_icon_from_file (GTK_WINDOW (main_window), PIXMAPDIR"/sound-juicer.png", NULL);
-
   select_all_menuitem = glade_xml_get_widget (glade, "select_all");
   deselect_all_menuitem = glade_xml_get_widget (glade, "deselect_all");
+  submit_menuitem = glade_xml_get_widget (glade, "submit");
   title_entry = glade_xml_get_widget (glade, "title_entry");
   artist_entry = glade_xml_get_widget (glade, "artist_entry");
   duration_label = glade_xml_get_widget (glade, "duration_label");
@@ -1283,9 +1382,6 @@ int main (int argc, char **argv)
     gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
   }
 
-  update_ui_for_album (NULL);
-  gtk_widget_show (main_window);
-
   extractor = SJ_EXTRACTOR (sj_extractor_new());
   error = sj_extractor_get_new_error (extractor);
   if (error) {
@@ -1293,11 +1389,15 @@ int main (int argc, char **argv)
     exit (1);
   }
 
+  update_ui_for_album (NULL);
+  gconf_bridge_bind_window_size(gconf_bridge_get(), GCONF_WINDOW, GTK_WINDOW (main_window));
+  gtk_widget_show (main_window);
+
   sj_play_init ();
 
   selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (track_listview));
   gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
-	  
+
   http_proxy_setup (gconf_client);
   baseuri_changed_cb (gconf_client, -1, gconf_client_get_entry (gconf_client, GCONF_BASEURI, NULL, TRUE, NULL), NULL);
   path_pattern_changed_cb (gconf_client, -1, gconf_client_get_entry (gconf_client, GCONF_PATH_PATTERN, NULL, TRUE, NULL), NULL);
@@ -1317,10 +1417,6 @@ int main (int argc, char **argv)
     return 0;
   }
 
-  /* Poke the CD drive every now and then */
-  tray_opened = nautilus_burn_drive_door_is_open (drive);
-  poll_id = g_timeout_add (2000, poll_tray_opened, NULL);
-
   gtk_main ();
   g_object_unref (metadata);
   g_object_unref (extractor);
